@@ -1,3 +1,21 @@
+/* 
+ * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
+ * 
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by the
+ * Free Software Foundation; either version 2 of the License, or (at your
+ * option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
+ */
+
 #include <ctype.h>
 #include <sys/stat.h>
 #include "jftpgw.h"
@@ -114,6 +132,7 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 	int chldpid =0;
 	const char* option = 0;
 	int nfd =0;
+	int hostnames2size;
 	int ahandle, shandle, ret, offset, port, i, maxfd, found;
 	struct sigaction sa;
 	struct sockaddr_in sin;
@@ -166,10 +185,10 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 	offset = 0;
 	/* we have to make the string suitable for quotstrtok by appending a
 	 * WHITESPACE character */
-	hostnames2 = (char*) malloc(strlen(hostnames) + 2);
+	hostnames2size = strlen(hostnames) + 2;
+	hostnames2 = (char*) malloc(hostnames2size);
 	enough_mem(hostnames2);
-	strcpy(hostnames2, hostnames);
-	strcat(hostnames2, " ");
+	snprintf(hostnames2, hostnames2size, "%s ", hostnames);
 	while ((part = quotstrtok(hostnames2, WHITESPACES, &offset))) {
 		portdel = strchr(part, ':');
 		*portdel = 0;
@@ -206,7 +225,9 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 	sa.sa_handler = childterm;
 	sig_chld_occurred = 0;
 	sigemptyset (&sa.sa_mask);
+#ifndef WINDOWS
 	sa.sa_flags = SA_RESTART;
+#endif
 	sigaction (SIGCHLD, &sa, 0);
 
 	found = 0;
@@ -280,7 +301,7 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 		}
 		peer_ip = getpeer_ip(ahandle);
 		log(7, "Connection from %s", peer_ip);
-		tg = get_showaddr(ahandle);
+		tg = get_showaddr(ahandle, SHOW_ORIGINAL_IP);
 		t_in.s_addr = tg;
 		log(7, "Tried to connect to %s", inet_ntoa(t_in));
 		ret = is_below_limit(inet_addr(peer_ip), &ls);
@@ -324,7 +345,9 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 			 */
 			sa.sa_handler = reap_chld_info;
 			sigemptyset(&sa.sa_mask);
+#ifndef WINDOWS
 			sa.sa_flags = SA_RESTART;
+#endif
 			sigaction(SIGCHLD, &sa, 0);
 
 			for (i = 0; i < clntinfo->boundsocket_niface; i++) {
@@ -351,7 +374,8 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 
 			clntinfo->clientsocket = ahandle;
 			clntinfo->dataport = get_usesport(ahandle) - 1;
-			clntinfo->clientaddr = get_showaddr(ahandle);
+			clntinfo->clientaddr = get_showaddr(ahandle,
+					SHOW_NOT_ORIGINAL_IP);
 			if (getuid() == 0 && geteuid() != 0) {
 				/* become root again to open the logfile(s) in
 				 * the logdir(s) */
@@ -399,7 +423,7 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 			if (getuid() == 0) {
 				log(8, "Changing ID back (init_logdirs(), "
 					"chroot())");
-				if (clntinfo->dataport > 1024) {
+				if (clntinfo->dataport > IPPORT_RESERVED) {
 					/* If we don't need root privileges
 					 * any more, change the UID */
 					changeid(get_option("runasuser"), UID);
@@ -410,14 +434,16 @@ int waitclient(const char* hostnames, struct clientinfo* clntinfo) {
 			if (get_option("welcomeline")) {
 				const char* welcome = get_option("welcomeline");
 				int i;
+				int welcomesize = strlen(welcome)+4+2+1;
 				char* welcmsg = (char*)
-					malloc(strlen(welcome)+4+2+1);
+					malloc(welcomesize);
 
 				enough_mem(welcmsg);
-				strcpy(welcmsg, "220 ");
-				strcat(welcmsg, welcome);
+				snprintf(welcmsg, welcomesize, "220 %s",
+						welcome);
 
-				/* remove all special characters */
+				/* remove all special characters, the length
+				 * of the text is not affected */
 				for (i = 0; i < strlen(welcmsg); i++) {
 					if (welcmsg[i] < 32) {
 						/* a special character */
