@@ -1,3 +1,21 @@
+/* 
+ * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
+ * 
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by the
+ * Free Software Foundation; either version 2 of the License, or (at your
+ * option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
+ */
+
 #include <fcntl.h>
 #include <ctype.h>
 #include <sys/time.h>
@@ -61,14 +79,17 @@ int handlecmds(const char* hostnames, struct clientinfo *clntinfo) {
 	enough_mem(li.host);
 
 	li.user = 0;
+	li.name = (char*) 0;
 
 	while (1) {
 contin:
 #ifdef SFTP_SUPPORT
 		if (clntinfo->servermode == SECURE_SFTP) {
+			li.service = "sftp";
 			cmdhandler = &secsftp_cmdhandler[0];
 		} else {
 #endif
+			li.service = "ftp";
 			cmdhandler = &std_cmdhandler[0];
 #ifdef SFTP_SUPPORT
 		}
@@ -86,7 +107,7 @@ contin:
 			li.respcode = 0;
 			li.transferred = 0;
 		}
-		buffer = ftp_readline(cs);
+		buffer = readline(cs);
 		if (!buffer && timeout) {
 			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
 			err_time_readline(cs);
@@ -163,6 +184,7 @@ contin:
 
 /* getuserdest() parses the "USER xyz" login string */
 
+#define IPBUFSIZE 16
 int getuserdest(char* buffer, struct clientinfo *clntinfo) {
 	char* dest, *user, *end, *orig_login;
 	char* buffnew = 0;
@@ -172,6 +194,10 @@ int getuserdest(char* buffer, struct clientinfo *clntinfo) {
 	f = 0;
 	cs = clntinfo->clientsocket;
 	clntinfo->servermode = getservermode();
+#ifdef NCFTP_TYPE5
+	log(9, "squeezing @");
+	char_squeeze(buffer, '@');
+#endif
 	while (buffer[i] == ' ')
 		i++;
 	if(i > strlen(buffer)) {
@@ -188,21 +214,21 @@ int getuserdest(char* buffer, struct clientinfo *clntinfo) {
 
 	opt = get_option("passallauth");
 	if (opt) {
+		int bufsize;
 		/* passallauth specified */
 		/* transform buffer into "buffer"@targethost,21 */
 
 		orig_login = strdup(&buffer[i]);
 		enough_mem(orig_login);
-		
-		buffer = (char*) realloc(buffer,
-				  strlen("USER ")
+		bufsize =  strlen("USER ")
 				+ 1                         /* " */
 				+ strlen(orig_login)        /* orig_login */
 				+ 1                         /* " */
 				+ 1                         /* @ */
 				+ strlen(opt)               /* opt */
-				+ 1);                       /* Termination */
-		sprintf(buffer, "USER \"%s\"@%s", orig_login, opt);
+				+ 1;                        /* Termination */
+		buffer = (char*) realloc(buffer, bufsize);
+		snprintf(buffer, bufsize, "USER \"%s\"@%s", orig_login, opt);
 		free(orig_login);
 		orig_login = 0;
 	}
@@ -254,7 +280,7 @@ int getuserdest(char* buffer, struct clientinfo *clntinfo) {
 	if (buffer[i-1] == '\0') {
 		const char* tmp;
 		const char* addtarget = 0;
-		char ipbuf[16];
+		char ipbuf[ IPBUFSIZE ];
 		/* first, lets terminate user. Since it is the last
 		 * value in the line, it is followed by an \r\n */
 		while(user[--f] < 32 && f >= 0) {
@@ -272,62 +298,65 @@ int getuserdest(char* buffer, struct clientinfo *clntinfo) {
         		namelen = sizeof(sin);
 
 			log(6, "No defaultroute specified, acting as transparent proxy");
-			sin.sin_addr.s_addr = get_showaddr(
-					clntinfo->clientsocket);
-			strcpy(ipbuf, inet_ntoa(*(struct in_addr *)
-						&sin.sin_addr));
+			sin.sin_addr.s_addr = get_showaddr(clntinfo->clientsocket, SHOW_ORIGINAL_IP);
+			strncpy(ipbuf, inet_ntoa(*(struct in_addr *)
+						&sin.sin_addr), IPBUFSIZE);
 			tmp = ipbuf;
 			
 			/* Determine the IP the client sees of us */
-		        if (getpeername(clntinfo->clientsocket,
+			if (getpeername(clntinfo->clientsocket,
 				(struct sockaddr *) &sin, &namelen) != 0) {
-        			log(3, "Could not get peername for TP");
+				log(3, "Could not get peername for TP");
 			}
 			clntinfo->clientaddr = get_our_addr_to(
 					sin.sin_addr.s_addr, sin.sin_port);
 			sin.sin_addr.s_addr = clntinfo->clientaddr;
 			log(8, "Our IP to the client is %s", inet_ntoa(sin.sin_addr));
 
- 			if ((fwhost = get_option("transparent-forward"))) {
- 				/* user => user@targethost */
- 				/* tmp => dest => forwardhost */
- 
- 				user = realloc(user, strlen(user) 
- 						+ 1 + strlen(ipbuf) + 1);
- 				enough_mem(user);
- 				clntinfo->user = user;
- 				strcat(user, "@");
- 				strcat(user, ipbuf);
- 
- 				addtarget = fwhost;
- 				log(8, "Using transparent proxy. Connecting to %s. User: %s", addtarget, user);
- 			} else {
- 				log(8, "Connecting to %s as transparent proxy",
- 					tmp);
- 				/* resize dest */
- 				dest = (char*) realloc(dest, strlen(tmp)+1);
- 				clntinfo->destination = dest;
- 				strcpy(dest, tmp);
- 				/* we don't have to free tmp */
- 				addtarget = 0;
- 			}
+			if ((fwhost = get_option("transparent-forward"))) {
+				/* user => user@targethost */
+				/* tmp => dest => forwardhost */
+
+				user = realloc(user, strlen(user) 
+						+ 1 + strlen(ipbuf) + 1);
+				enough_mem(user);
+				clntinfo->user = user;
+				strcat(user, "@");
+				strcat(user, ipbuf);
+
+				addtarget = fwhost;
+				log(8, "Using transparent proxy. Connecting to %s. User: %s", addtarget, user);
+			} else {
+				int destsize;
+				log(8, "Connecting to %s as transparent proxy",
+					tmp);
+				/* resize dest */
+				destsize = strlen(tmp) + 1;
+				dest = (char*) realloc(dest, destsize);
+				enough_mem(dest);
+				clntinfo->destination = dest;
+				snprintf(dest, destsize, "%s", tmp);
+				/* we don't have to free tmp */
+				addtarget = 0;
+			}
 		} else {
 			addtarget = tmp;
 		}
 		log(9, "Addtarget: %s", addtarget);
 		if (addtarget) {
 			/* copy the option of defaultforward into buffer */
-			buffnew = (char*) malloc(strlen(buffer) 
- 							+ strlen(addtarget)
- 							+ 1 + 1);
+			int buffnewsize = strlen(buffer)
+						+ strlen(addtarget)
+						+ 1
+						+ 1;
+			buffnew = (char*) malloc(buffnewsize);
 			enough_mem(buffnew);
 			dest = (char*) realloc(dest, strlen(addtarget) + 1);
- 			clntinfo->destination = dest;
-			strcpy(buffnew, buffer);
-			strcat(buffnew, "@");
- 			strcat(buffnew, addtarget);
+			clntinfo->destination = dest;
+			snprintf(buffnew, buffnewsize, "%s@%s",
+					buffer, addtarget);
 			buffer = buffnew;
- 			/* we don't have to free addtarget */
+			/* we don't have to free addtarget */
 			log(9, "Appended defaultforward option. New login string: %s. buffer[i]: %s", buffer, buffer+i);
 		}
 	}
@@ -406,7 +435,6 @@ int getuserdest(char* buffer, struct clientinfo *clntinfo) {
 	if (!clntinfo->destinationport) {
 		clntinfo->destinationport = getserverport();
 	}
-
 	log(7, "Client logged in: User: %s, Dest: %s", user, dest);
 	switch (clntinfo->servermode) {
 		case SECURE_SFTP:
@@ -430,8 +458,16 @@ int getpasswd(char* buffer, struct clientinfo *clntinfo) {
 	
 	f = strlen("PASS ");
 	i = 0;
-	while (buffer[f] == ' ')
+	while (buffer[f] == ' ') {
 		f++;
+	}
+#ifdef NCFTP_TYPE5
+	log(9, "chopping first @ in password: %s", &buffer[f]);
+	while (buffer[f] == '@') {
+		f++;
+	}
+	log(9, "after chopping: %s", &buffer[f]);
+#endif
 	while (buffer[f]) {
 		pass[i] = buffer[f];
 		i++; f++;
@@ -452,12 +488,15 @@ int passcmd(char* buffer, struct clientinfo *clntinfo) {
 	char* sendbuf =0;
 	char *last = 0;
 	int ret;
+	int sendbufsize;
+	int transfer_start;
 
-	sendbuf = (char*) malloc(strlen(buffer) + 3);
-	strcpy(sendbuf, buffer);
-	strcat(sendbuf, "\r\n");
+	sendbufsize = strlen(buffer) + 3;
+	sendbuf = (char*) malloc(sendbufsize);
+	snprintf(sendbuf, sendbufsize, "%s\r\n", buffer);
 	say(ss, sendbuf);
 	free(sendbuf);
+	li.complete = 0;
 	last = passall(ss, cs);
 	if (!last) {
 		if (timeout) {
@@ -483,8 +522,10 @@ int passcmd(char* buffer, struct clientinfo *clntinfo) {
 			free(last);
 			return -1;
 		}
-		
+
+		transfer_start = time(NULL);
 		ret = transmitfile(clntinfo);
+		li.transfer_duration = time(NULL) - transfer_start;
 
 		if (!ret) {
 			char* t = passall(ss, cs);
@@ -507,6 +548,7 @@ int passcmd(char* buffer, struct clientinfo *clntinfo) {
 	}
 	free(/*at*/ last);  /*  ;-)  */
 	last = 0;
+	li.complete = li.respcode == 226;
 	return 0;
 }
 
@@ -609,8 +651,11 @@ int initiate_transfer(struct clientinfo *clntinfo) {
 	return 0;
 }
 
-/* #define TRANSMITBUFSIZE  (10*1024) */
+#ifdef WINDOWS
+#define TRANSMITBUFSIZE  (10*1024)
+#else
 #define TRANSMITBUFSIZE  PIPE_BUF
+#endif
 int transmitfile(struct clientinfo *clntinfo)  {
 	char* buffer = (char*) malloc(TRANSMITBUFSIZE);
 	char* pbuf = 0;
@@ -659,11 +704,14 @@ int transmitfile(struct clientinfo *clntinfo)  {
 	if (ret < 0) {
 		log(3, "sigprocmask() error: %s", strerror(errno));
 	}
-	
+
+	/* read from the clientsocket as well as from the dataserversocket */
+	/* no writeset is specified here. The client could send an abort (it
+	 * is also in exceptset) and the server sends data */
 	while((sret = select(MAX_VAL(clntinfo->dataserversock, cs) + 1,
 				&readset, NULL, &exceptset,
 				&readtime) > 0)) {
-		/* save the errno value from sigprocmask() */
+		/* save the errno value of select from sigprocmask() */
 		n = errno;
 		ret = sigprocmask(SIG_UNBLOCK, &sigset, &oldset);
 		if (ret < 0) {
@@ -672,18 +720,25 @@ int transmitfile(struct clientinfo *clntinfo)  {
 		}
 		errno = n;
 
+		/* Can we read data from the client ? */
 		if (FD_ISSET(clntinfo->dataserversock, &readset)) {
 			count = read(clntinfo->dataserversock,
 				buffer, TRANSMITBUFSIZE);
 			if (count == 0) {
+				log(8, "Read 0 bytes at %s (%d)", __FILE__, __LINE__);
 				break;
 			}
 			if (count < 0) {
 				log(3, "read error: %s", strerror(errno));
+				/* comm */
 				sret = 1;
 				error = 1;
 				break;
 			}
+			/* This is only needed for sftp mode connections. If
+			 * the proxy acts between a real FTP client and a
+			 * real FTP server, both will do the job of
+			 * converting */
 			if (clntinfo->havetoconvert != CONV_NOTCONVERT) {
 				char* tmp;
 				if (clntinfo->havetoconvert == CONV_TOASCII) {
@@ -701,8 +756,10 @@ int transmitfile(struct clientinfo *clntinfo)  {
 				}
 			}
 
+			/* comm */
 			pbuf = buffer;
 			do {
+				/* now write all the read data */
 				FD_ZERO(&writeset);
 				FD_SET(clntinfo->dataclientsock, &writeset);
 
@@ -715,9 +772,11 @@ int transmitfile(struct clientinfo *clntinfo)  {
 						strerror(errno));
 				}
 
+				/* selecting to write */
 				scret = select(clntinfo->dataclientsock + 1,
 					NULL, &writeset, NULL, &writetime);
-				/* save the errno value from sigprocmask() */
+				/* save the errno value of select from
+				 * sigprocmask() */
 				n = errno;
 				ret = sigprocmask(SIG_UNBLOCK, &sigset,
 						&oldset);
@@ -733,16 +792,28 @@ int transmitfile(struct clientinfo *clntinfo)  {
 				}
 				/* otherwise the descriptor must be ready */
 				nwritten = write(clntinfo->dataclientsock,
-						buffer, count);
-				if(nwritten < 0) break;
+						pbuf, count);
+				if (nwritten < 0) {
+					log(8, "Error writing (%s, %d): %s",
+							__FILE__, __LINE__,
+							strerror(errno));
+					break;
+				}
+				if (nwritten == 0) {
+					log(8, "Wrote 0 bytes (%s, %d)",
+							__FILE__, __LINE__);
+				}
 				totwritten += nwritten;
-				done = time(NULL);
-				delay = (totwritten/1024.00) /
-					clntinfo->throughput;
-				delay += start;
-				delay -= done;
-				if (delay > 0 && clntinfo->throughput != -1) {
-					sleep(delay);
+				/* calculate the delay time */
+				if (clntinfo->throughput >= 0) {
+					done = time(NULL);
+					delay = (totwritten / 1024.00) /
+						clntinfo->throughput;
+					delay += start;
+					delay -= done;
+					if (delay > 0) {
+						sleep(delay);
+					}
 				}
 				if (nwritten != count) {
 					pbuf += nwritten;
@@ -752,7 +823,7 @@ int transmitfile(struct clientinfo *clntinfo)  {
 				}
 			} while (pbuf);
 		}
-		if(FD_ISSET(cs, &exceptset) || FD_ISSET(cs, &readset)) {
+		if (FD_ISSET(cs, &exceptset) || FD_ISSET(cs, &readset)) {
 			if (checkforabort(clntinfo)) {
 				error = 2;
 				break;
@@ -762,10 +833,10 @@ int transmitfile(struct clientinfo *clntinfo)  {
 		FD_ZERO(&readset);
 		FD_ZERO(&writeset);
 		FD_ZERO(&exceptset);
-	        FD_SET(clntinfo->dataserversock, &readset);
-	        FD_SET(clntinfo->dataclientsock, &writeset);
-	        FD_SET(cs, &exceptset);
-	        FD_SET(cs, &readset);
+		FD_SET(clntinfo->dataserversock, &readset);
+		FD_SET(clntinfo->dataclientsock, &writeset);
+		FD_SET(cs, &exceptset);
+		FD_SET(cs, &readset);
 		/* Linux modifies the timeout struct */
 		readtime.tv_sec = 300;
 		readtime.tv_usec = 0;
@@ -806,10 +877,6 @@ int transmitfile(struct clientinfo *clntinfo)  {
 	close(clntinfo->dataserversock);
 	clntinfo->dataclientsock = -1;
 	clntinfo->dataserversock = -1;
-	/*
-	clntinfo->servermode = ASCLIENT;
-	clntinfo->clientmode = ASCLIENT;
-	*/
 
 	li.transferred = totwritten;
 	log(7, "Transferred %d bytes", totwritten);
@@ -824,19 +891,26 @@ int checkforabort(struct clientinfo* clntinfo) {
 	int serverfd = clntinfo->serversocket;
 	char buffer[1024];
 	char* buf =0;
-	char rabseq[3] = { IAC, IP, 0 };
-	char sabseq[4] = { IAC, IP, IAC, 0 };
-	char dmseq[1] = { DM };
+	char rabseq[3] = { (char) IAC, (char) IP, (char) 0 };
+	char sabseq[4] = { (char) IAC, (char) IP, (char) IAC, (char) 0 };
+	char dmseq[1]  = { (char) DM };
 	int ret;
 
 	struct timeval tm = { 300, 0 };
 	fd_set set, exset;
-	FD_ZERO(&set);
-	FD_SET(clientfd, &set);
-	
+
 	/* this read won't block */
 	ret = read(clientfd, buffer, sizeof(buffer) - 1);
 	if (ret < 0) {
+		/* return, if clientfd is negative for example. "Not a valid
+		 * file descriptor". This is important since FD_SET might
+		 * segfault with a negative filedescriptor id */
+		return 0;
+	}
+	if (ret == 0) {
+		log(8, "Read 0 bytes at %s, %d", __FILE__, __LINE__);
+		log(8, "Read from clientfd, closing it! Should not happen!");
+		close(clientfd);
 		return 0;
 	}
 	buffer[ret] = '\0';
@@ -845,6 +919,7 @@ int checkforabort(struct clientinfo* clntinfo) {
 		return 0;
 	}
 
+	FD_ZERO(&set);
 	FD_SET(clientfd, &set);
 	ret = select(clientfd + 1, &set, NULL, &set, &tm);
 
@@ -853,13 +928,22 @@ int checkforabort(struct clientinfo* clntinfo) {
 	}
 
 	ret = read(clientfd, buffer, sizeof(buffer) - 1);
+	if (ret < 0) {
+		return 0;
+	}
+	if (ret == 0) {
+		log(8, "Read 0 bytes at %s, %d", __FILE__, __LINE__);
+		log(8, "Read from clientfd, closing it! Should not happen!");
+		close(clientfd);
+		return 0;
+	}
 	buffer[ret] = '\0';
 
 	if(buffer[0] != (char) DM ||
 		strncasecmp(buffer + 1,"ABOR",4) != 0) {
 		return 0;
 	}
-		
+
 	/* abort */
 
 	if (clntinfo->dataclientsock == -1) {
@@ -873,7 +957,7 @@ int checkforabort(struct clientinfo* clntinfo) {
 				"Closed data connection.\r\n");
 		say(clientfd, "226 Abort successful\r\n");
 	}
-	
+
 	if (clntinfo->dataserversock == -1) {
 		/* if we don't have a connection to the server we can
 		 * quit here
@@ -889,14 +973,14 @@ int checkforabort(struct clientinfo* clntinfo) {
 		return 1;
 	}
 #endif
-	
+
 	if (send(serverfd, sabseq, 3, MSG_OOB) != 3
 		|| send(serverfd, dmseq, 1, 0) != 1) {
 		log(3, "send() error: %s", strerror(errno));
 	}
 	say(serverfd, "ABOR\r\n");
 
-	
+
 	FD_ZERO(&set);
 	FD_ZERO(&exset);
 	FD_SET(serverfd, &set);
