@@ -1,7 +1,26 @@
+/* 
+ * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
+ * 
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by the
+ * Free Software Foundation; either version 2 of the License, or (at your
+ * option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
+ */
+
 #include <ctype.h>
 #include <stdlib.h>
 #include <limits.h>
 #include <pwd.h>
+#include <syslog.h>
 #include "jftpgw.h"
 
 char* valid_keys[] = { "accessfile", "serverport", "defaultmode",
@@ -10,7 +29,9 @@ char* valid_keys[] = { "accessfile", "serverport", "defaultmode",
 			"atstartchangerootdir", "changerootdir", "limit",
 			"ssh-passwd", "passallauth", "ssh-options",
 			"transparent-forward", "pidfile", "welcomeline",
-			"listen", "debuglevel", (char*) 0 };
+			"listen", "debuglevel", "allowreservedports",
+			"allowforeignaddress", "passiveportrange",
+			"activeportrange", "logstyle", (char*) 0 };
 
 struct optionstruct {
 	char* key;
@@ -52,6 +73,11 @@ struct sshoptionsstruct {
 	struct sshoptionsstruct* next;
 } *sshoptionsbase, *sshoptionscur;
 
+struct portrangestruct *actvrangebase, *pasvrangebase,
+			*actvrangecur, *pasvrangecur;
+unsigned int actvportsamount;
+unsigned int pasvportsamount;
+
 struct lchar {
 	char* entry;
 	struct lchar* next;
@@ -61,6 +87,9 @@ struct lchar {
 extern struct configstruct config;
 extern struct uidstruct runasuser;
 
+/* from jftpgw.c */
+extern char* conffilename;
+
 struct loginstruct lsforward;
 struct limitstruct *limitbase, *limitcur;
 
@@ -71,14 +100,24 @@ int add_throughput(const char*);
 int add_limit(const char*);
 int add_sshpasswd(const char*);
 int add_sshoptions(const char*);
+int add_rangespec(const char*);
 int match_username(const char*, const char*);
 int match_addrs(unsigned long int, const char*, unsigned long int,
                 unsigned long int, const char*);
 int save_runasuser_uid(void);
+void free_llistst(struct lchar*);
 
 int conf_delete_registered = 0;
 
 
+
+int set_conffilename(const char* arg) {
+	free(conffilename);
+	conffilename = strdup(arg);
+	enough_mem(conffilename);
+	return 0;
+}
+
 int read_config(const char* filename) {
 	int ret;
 	FILE* conffile = open_file(filename);
@@ -98,6 +137,8 @@ int read_config(const char* filename) {
 	limitbase = limitcur = (struct limitstruct*) 0;
 	sshpasswdbase = sshpasswdcur = (struct sshpasswdstruct*) 0;
 	sshoptionsbase = sshoptionscur = (struct sshoptionsstruct*) 0;
+	pasvrangebase = pasvrangecur = (struct portrangestruct*) 0;
+	actvrangebase = actvrangecur = (struct portrangestruct*) 0;
 	ret = config_read_options(conffile);
 	if (ret) {
 		return -1;
@@ -194,6 +235,7 @@ int config_read_options(FILE* file) {
 #define LIMIT        3
 #define SSHPASSWD    4
 #define SSHOPTIONS   5
+#define RANGESPEC    6
 
 	int entrytype =0;
 
@@ -225,6 +267,12 @@ int config_read_options(FILE* file) {
 		if (strncmp(entry, "ssh-options", strlen("ssh-options")) == 0) {
 			entrytype = SSHOPTIONS;
 		}
+		if (strncmp(entry, "passiveportrange",
+			    strlen("passiveportrange")) == 0
+		   || strncmp(entry, "activeportrange",
+			      strlen("activeportrange")) == 0) {
+			entrytype = RANGESPEC;
+		}
 		switch(entrytype) {
 			case FORWARD:
 				ret = add_forward(entry);
@@ -241,6 +289,9 @@ int config_read_options(FILE* file) {
 			case SSHOPTIONS:
 				ret = add_sshoptions(entry);
 				break;
+			case RANGESPEC:
+				ret = add_rangespec(entry);
+				break;
 			default:
 				ret = add_to_options(entry);
 		}
@@ -263,6 +314,7 @@ int add_to_options(char* entry) {
 	if (!*space) {
 		return -1;
 	} else {
+		int valuesize;
 		ptr = (struct optionstruct*)
 				malloc(sizeof(struct optionstruct));
 		enough_mem(ptr);
@@ -292,11 +344,12 @@ int add_to_options(char* entry) {
 		if (strstr(optionscur->key, "specs")) {
 			add = 1;
 		}
-		optionscur->value = (char*) malloc(strlen(p) + 1 + add);
+		valuesize = strlen(p) + 1 + add;
+		optionscur->value = (char*) malloc(valuesize);
 		if (add) {
 			optionscur->value[0] = ' ';
 		}
-		strcpy(optionscur->value + add, p);
+		strncpy(optionscur->value + add, p, valuesize - add);
 
 		while (optionscur->value[strlen(optionscur->value)-1] == '\n') {
 			optionscur->value[strlen(optionscur->value)-1] = '\0';
@@ -352,10 +405,13 @@ int is_valid_key(const char* keyval) {
 char* quotstrtok(const char* s, const char* delim, int *past_offset) {
 
 	int i = *past_offset;
-	char* q = 0, *ret =0, *r =0;
+	const char* q = 0;
+	char *ret =0, *r =0;
 	const char* ubound, *lbound;
 
-	while(s[i] && isspace((int)s[i])) {
+	/*while(s[i] && isspace((int)s[i])) {*/
+	/* move forth if s[i] is one of the delimiters */
+	while(s[i] && strchr(delim, (int)s[i])) {
 		i++;
 	}
 	if (s[i] == 0)
@@ -367,7 +423,11 @@ char* quotstrtok(const char* s, const char* delim, int *past_offset) {
 		q = strpbrk(&s[i], delim);
 	}
 	if (!q || !*q) {
-		return 0;
+		if (strlen(&s[i]) == 0) {
+			return 0;
+		} else {
+			q = s + i + strlen(&s[i]);
+		}
 	}
 	/* chop quotation marks on both sides */
 	if (s[i] == '"' && *(q-1) == '"') {
@@ -388,7 +448,7 @@ char* quotstrtok(const char* s, const char* delim, int *past_offset) {
 		r = ret + strlen(ret) - 1;
 		while (iscntrl((int)*r) && strlen(ret)) {
 			*r = '\0';
-			r--;
+		r--;
 		}
 	}
 
@@ -458,32 +518,28 @@ int get_forward(const char* user, struct loginstruct* logininfo) {
 
 	forwardcur = forwardbase;
 	do {
-		if (strcmp(user, forwardcur->fwlogin) == 0) {
-			logininfo->fwlogin = 
-				malloc (strlen(forwardcur->fwlogin) + 1);
+		if (strcmp(user, forwardcur->fwlogin) == 0
+		/* a "*" matches all usernames */
+		 || strcmp("*", forwardcur->fwlogin) == 0) {
+			logininfo->fwlogin =
+				strdup(forwardcur->fwlogin);
 			enough_mem(logininfo->fwlogin);
-			strcpy(logininfo->fwlogin, forwardcur->fwlogin);
 
-			logininfo->login = 
-				malloc (strlen(forwardcur->destlogin) + 1);
+			logininfo->login =
+				strdup(forwardcur->destlogin);
 			enough_mem(logininfo->login);
-			strcpy(logininfo->login, forwardcur->destlogin);
-
 
 			if (forwardcur->fwpass) {
-				logininfo->fwpass = 
-				malloc (strlen(forwardcur->fwpass) + 1);
+				logininfo->fwpass = strdup(forwardcur->fwpass);
 				enough_mem(logininfo->fwpass);
-				strcpy(logininfo->fwpass, forwardcur->fwpass);
 			} else {
 				logininfo->fwpass = 0;
 			}
 
 			if (forwardcur->destpass) {
 				logininfo->pass =
-				malloc (strlen(forwardcur->destpass) + 1);
+					strdup(forwardcur->destpass);
 				enough_mem(logininfo->pass);
-				strcpy(logininfo->pass, forwardcur->destpass);
 			} else {
 				logininfo->pass = 0;
 			}
@@ -534,30 +590,13 @@ int checkfwpass(struct clientinfo* clntinfo) {
 		clntinfo->pass = realloc(clntinfo->pass,
 					strlen(lsforward.pass) + 1);
 		enough_mem(clntinfo->pass);
-		strcpy(clntinfo->pass, lsforward.pass);
+		strncpy(clntinfo->pass, lsforward.pass,
+				strlen(lsforward.pass) /* always true */);
 		return 1;
 	}
 	return 0;
 }
 
-unsigned long int get_showaddr(int shandle) {
-	struct sockaddr_in sin;
-	int i;
-#ifdef HAVE_SOCKLEN_T
-	socklen_t slen;
-#else
-	int slen;
-#endif
-	slen = sizeof (sin);
-
-	i = getsockname(shandle, (struct sockaddr*) &sin, &slen);
-	if (i != 0) {
-		log(2, "getsockname failed. Can't get the IP of the interface");
-		return -1;
-	}
-	return sin.sin_addr.s_addr;
-}
-
 
 
 int add_throughput(const char* entry) {
@@ -642,11 +681,11 @@ float get_rate(const char *fwuser, const char* user,
 	}
 
 	if (user) {
+		int arrusersize = strlen(user) + 2;
 		/* user may be (char*) 0 if no forward is used */
-		arruser = malloc(strlen(user) + 2);
+		arruser = malloc(arrusersize);
 		enough_mem(arruser);
-		arruser[0] = '>';
-		strcpy(&arruser[1], user);
+		snprintf(arruser, arrusersize, ">%s", user);
 	} else {
 		arruser = (char*) 0;
 	}
@@ -740,7 +779,7 @@ int match_addrs(unsigned long int ip, const char* name,
 					return 1;
 				}
 				i++;
-                	} while (host->h_addr_list[i]);
+			} while (host->h_addr_list[i]);
 		}
 	}
 
@@ -827,7 +866,7 @@ int add_limit(const char* entry) {
 	/* the last entry must be the number of connects allowed */
 
 	connamount = strtol(llistcur->entry, NULL, 10);
-	if (errno == ERANGE  
+	if (errno == ERANGE
 			&& (connamount == LONG_MIN || connamount == LONG_MAX)) {
 		log(3, "Error reading the amount of allowed connections in %s",
 			entry);
@@ -1067,6 +1106,195 @@ int add_sshoptions(const char* entry) {
 }
 
 
+int add_rangespec(const char* entry) {
+	int offset = 0;
+	unsigned int *portsamount;
+	struct portrangestruct** prsbase, **prscur, *prs;
+	char* startportstr, *endportstr;
+	unsigned int startport, endport;
+	char* typestr = quotstrtok(entry, WHITESPACES"\n", &offset);
+	struct lchar* llistbase = split_line(entry, typestr);
+	struct lchar* llistcur = llistbase;
+	struct lchar* prevtmp;
+
+	if (strcasecmp(typestr, "passiveportrange") == 0) {
+		prsbase = &pasvrangebase;
+		prscur  = &pasvrangecur;
+		portsamount = &pasvportsamount;
+	} else {
+		prsbase = &actvrangebase;
+		prscur  = &actvrangecur;
+		portsamount = &actvportsamount;
+	}
+	free(typestr);
+	typestr = (char*) 0;
+
+	while (llistcur) {
+		offset = 0;
+		startportstr = quotstrtok(llistcur->entry, ":", &offset);
+		endportstr = quotstrtok(llistcur->entry, ":", &offset);
+		if (!startportstr || !endportstr) {
+			log(3, "invalid port range: %s", llistcur->entry);
+			/* free */
+			free_llistst(llistbase);
+			return -1;
+		}
+		if (!*startportstr || !*endportstr) {
+			log(3, "invalid port range: %s", llistcur->entry);
+			/* free */
+			free_llistst(llistbase);
+			return -1;
+		}
+		startport = strtol(startportstr, NULL, 10);
+		if (errno == ERANGE
+				&& (startport == LONG_MIN || startport == LONG_MAX)) {
+			log(3, "Error reading the starting port number in %s",
+				startportstr);
+			free_llistst(llistbase);
+			free(startportstr);
+			return -1;
+		}
+		free(startportstr);
+		startportstr = (char*) 0;
+
+		endport = strtol(endportstr, NULL, 10);
+		if (errno == ERANGE
+				&& (endport == LONG_MIN || endport == LONG_MAX)) {
+			log(3, "Error reading the ending port number in %s",
+				endportstr);
+			free_llistst(llistbase);
+			free(endportstr);
+			return -1;
+		}
+		free(endportstr);
+		endportstr = (char*) 0;
+
+		/* startport really below or equal to endport ? */
+		if (startport > endport) {
+			log(4, "Port range %d:%d invalid (starting port number"
+				" %d is _above_ endport %d)",
+					startport, endport,
+					startport, endport);
+			free_llistst(llistbase);
+			return -1;
+		}
+
+		/* both must be between (inclusive) 1024 and 65535 */
+		if (startport > 65535   ||   startport < /* 1024 */ 0
+		   || endport > 65535   ||   endport   < /* 1024 */ 0) {
+
+			log(4, "Port range %d:%d invalid (at least one port "
+				"number not in range 1024:65535)",
+					startport, endport);
+			free_llistst(llistbase);
+			return -1;
+		}
+
+		/* I want only to allow clear ranges. This is better for the
+		 * checking, for the selection of a port and the admin is
+		 * forced to specify his ports in a clear manner  :-)
+		 *
+		 * That means that a portrange like:
+		 *
+		 * 1000:2000   1500:1600   is invalid, since 1500:1600 is
+		 * already in 1000:2000
+		 *
+		 * Also invalid:
+		 *
+		 * 1000:2000   1800:2200   (200 portnumbers are the same)
+		 * 1000:2000   2000:3000   (1 portnumber is the same)
+		 * 1000:2000   500:1000    (1 portnumber is the same)
+		 * 1000:2000   799:1200    (200 portnumbers are the same)
+		 *
+		 * criteria: startport and endport => let's say just pno
+		 *
+		 * If pno is between (or equal to) the range of another port
+		 * specification, state that is invalid.
+		 * ------
+		 * What happens if we first specify
+		 *
+		 * 300:400
+		 *
+		 * and then
+		 *
+		 * 200:1000 ?
+		 *
+		 * We have to do the checking for _each_ new range to _all_
+		 * ranges. Compare every range to every other range
+		 *
+		 */
+
+		prs = *prsbase;
+		while (prs) {
+			if (
+			    (startport >= prs->startport &&
+			     startport <= prs->endport)
+			    ||
+			    (endport >= prs->startport &&
+			     endport <= prs->endport)
+			   ) {
+				log(4, "Port range %d:%d invalid (parts "
+					"covered by another range (%d:%d))",
+						startport, endport,
+						prs->startport, prs->endport);
+				free_llistst(llistbase);
+				return -1;
+			}
+
+			if (
+			    (prs->startport >= startport &&
+			     prs->startport <= endport)
+			    ||
+			    (prs->endport >= startport &&
+			     prs->endport <= endport)
+			    ) {
+				log(4, "Port range %d:%d invalid (parts "
+					"covered by another range (%d:%d))",
+						prs->startport, prs->endport,
+						startport, endport);
+				free_llistst(llistbase);
+				return -1;
+			}
+			prs = prs->next;
+		}
+
+		prs = (struct portrangestruct*)
+				malloc(sizeof(struct portrangestruct));
+		if (*prscur) {
+			(*prscur)->next = prs;
+		}
+
+		*prscur = prs;
+
+		if (!*prsbase) {
+			/* the first one */
+			*prsbase = *prscur;
+			*portsamount = 0;
+		}
+		(*prscur)->startport = startport;
+		(*prscur)->endport = endport;
+		(*prscur)->next = (struct portrangestruct*) 0;
+
+		/* startport: 7
+		 * endport:  10
+		 *
+		 * We have    10 - 7 + 1    =    3 + 1    =    4      ports
+		 *
+		 * 7, 8, 9, 10
+		 * */
+		*portsamount += (endport - startport) + 1;
+
+		/* iterate through and always chop off the first element */
+		prevtmp = llistcur;
+		llistcur = llistcur->next;
+		llistbase = llistcur;
+		free(prevtmp->entry);
+		free(prevtmp);
+		prevtmp = 0;
+	}
+
+	return 0;
+}
 
 
 int is_below_limit(unsigned long int iaddr, struct limitstruct** ls) {
@@ -1391,17 +1619,15 @@ const char* getaccessfile() {
 	const char* opt = get_option("accessfile");
 	if (!opt) {
 		char *b;
+		int bsize;
 		fprintf(stderr, "No accessfile option specified. Assuming "
 				"%s as default\n", DEFAULTACCESSFILE);
 		log(4, "No accessfile option specified. Assuming "
 		    "%s as default", DEFAULTACCESSFILE);
-		b = (char*) malloc(strlen("accessfile") 
-				+ strlen(DEFAULTACCESSFILE) + 3);
+		bsize = strlen("accessfile") + strlen(DEFAULTACCESSFILE) + 3;
+		b = (char*) malloc(bsize);
 		enough_mem(b);
-		strcpy(b, "accessfile");
-		strcat(b, " ");
-		strcat(b, DEFAULTACCESSFILE);
-		strcat(b, "\n");
+		snprintf(b, bsize, "accessfile %s\n", DEFAULTACCESSFILE);
 		add_to_options(b);
 		free(b);
 		return DEFAULTACCESSFILE;
@@ -1507,7 +1733,21 @@ void free_sshoptionsst(struct sshoptionsstruct* ps) {
 	free(ps);
 }
 
+void free_portrangest(struct portrangestruct* prs) {
+	if (!prs) {
+		return;
+	}
+	free_portrangest(prs->next);
+	free(prs);
+}
 
+void free_llistst(struct lchar* ls) {
+	if (!ls) {
+		return;
+	}
+	free_llistst(ls->next);
+	free(ls);
+}
 
 void reset_confst(struct configstruct* cs) {
 	if (!cs) {
@@ -1522,6 +1762,9 @@ void reset_confst(struct configstruct* cs) {
 		fclose(cs->logf);
 		cs->logf = (FILE*) 0;
 	}
+	if (cs->syslog) {
+		closelog();
+	}
 	free_cmdlogentst(&cs->cmdlogfiles, &cs->cmdlogfiles);
 	free_cmdlogentst(&cs->cmdlogdirs, &cs->cmdlogdirs);
 	cs->cmdlogfiles.next = (struct cmdlogent_t*) 0;
@@ -1535,7 +1778,9 @@ void conf_delete_config(void) {
 	free_limitst(limitbase);
 	free_sshpasswdst(sshpasswdbase);
 	free_sshoptionsst(sshoptionsbase);
-	reset_confst(&config);
+	free_portrangest(actvrangebase);
+	free_portrangest(pasvrangebase);
+	/* reset_confst(&config);  - moved to jftpgw.c */
 	if (runasuser.username) {
 		free(runasuser.username);
 		runasuser.username = 0;
@@ -1546,5 +1791,7 @@ void conf_delete_config(void) {
 	limitbase = (struct limitstruct*) 0;
 	sshpasswdbase = (struct sshpasswdstruct*) 0;
 	sshoptionsbase = (struct sshoptionsstruct*) 0;
+	actvrangebase = (struct portrangestruct*) 0;
+	pasvrangebase = (struct portrangestruct*) 0;
 }
 
