@@ -1,5 +1,24 @@
+/* 
+ * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
+ * 
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by the
+ * Free Software Foundation; either version 2 of the License, or (at your
+ * option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
+ */
+
 #include <sys/types.h>
 #include <sys/time.h>
+#include <time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <netdb.h>
@@ -16,6 +35,8 @@
 #include <limits.h>  /* for UINT_MAX -> OSF does not accept inet_addr() == -1 */
 #include "log.h"
 
+/* #define NCFTP_TYPE5 */
+
 /* include the autoheader file config.h */
 
 #ifdef HAVE_CONFIG_H
@@ -23,6 +44,7 @@
 #endif
 
 #define MAX_VAL(a,b) ((a)>(b)?(a):(b))
+#define MIN_VAL(a,b) ((a)<(b)?(a):(b))
 
 #define DEFAULTCONFFILE		CONFFILE
 #define DEFAULTBINDPORT		2370
@@ -59,6 +81,15 @@
 
 #define MAX_LINE_SIZE		4096
 
+#ifndef INPORT_ANY
+#define INPORT_ANY		0
+#endif
+
+#if !defined(HAVE_SNPRINTF) || !defined(HAVE_VSNPRINTF)
+  int snprintf (char *str, size_t count, const char *fmt, ...);
+  int vsnprintf (char *str, size_t count, const char *fmt, va_list arg);
+#endif
+
 
 extern FILE* logfile;
 struct allowed_lst;
@@ -85,6 +116,9 @@ struct clientinfo {
 	float throughput;
 };
 
+#define LS_COMMONLOG		0	/* common log file style */
+#define LS_XFERLOG		1	/* xferlog style */
+
 struct configstruct {
 	int syslog;
 	int debuglevel;
@@ -96,6 +130,7 @@ struct configstruct {
 		char* specs;
 		FILE* logf;
 		struct cmdlogent_t* next;
+		int style;
 	} cmdlogfiles, cmdlogdirs;
 };
 
@@ -139,6 +174,13 @@ struct limitstruct {
 	struct limitstruct* next;
 };
 
+struct portrangestruct {
+	unsigned int startport;
+	unsigned int endport;
+	struct portrangestruct* next;
+};
+
+
 /* signal handling functions */
 void read_default_conf(int);
 void childterm(int);
@@ -159,6 +201,7 @@ void removepidfile(void);
 int pasvclient(struct clientinfo*);
 int pasvserver(struct clientinfo*);
 
+int portcommandcheck(char*, struct sockaddr_in*, struct clientinfo*);
 int activeclient(char*, struct clientinfo*);
 int activeserver(char**, struct clientinfo*);
 
@@ -168,12 +211,15 @@ int sftp_open(char**, int);
 
 int checkdigits(const char*, const int);
 char* get_ip(int);
-unsigned long int get_showaddr(int);
+#define SHOW_ORIGINAL_IP		0
+#define SHOW_NOT_ORIGINAL_IP		1
+unsigned long int get_showaddr(int, int);
 int waitclient(const char*, struct clientinfo*);
 int handlecmds(const char*, struct clientinfo*);
 int say(int, const char*);
 int passcmd(char*, struct clientinfo*);
-int openownport(struct sockaddr_in *, struct clientinfo *, int);
+int openlocalport(struct sockaddr_in *, struct clientinfo *,
+		  struct portrangestruct *, int);
 int openportiaddr(unsigned long, unsigned int,
 		  unsigned int, struct clientinfo *);
 int openportaddr(char*, unsigned int,
@@ -188,6 +234,7 @@ int transmitfile(struct clientinfo *);
 int initiate_transfer(struct clientinfo *);
 
 char* ftp_readline(int);
+char* readline(int);
 struct message readall(int);
 int ftp_getrc(int, char**);
 char* passall(int, int);
@@ -198,6 +245,7 @@ void closedescriptors(void);
 int daemonize(void);
 void clean_allowedlist(void);
 void enough_mem(const void*);
+void scnprintf (char *os, size_t len, const char *str, ...);
 unsigned long int setlastbits(int);
 int cmp_domains(const char* name, const char* pattern);
 void err_time_readline(int fd);
@@ -212,7 +260,8 @@ int respcode(const char*);
 int init_logdirs(const char*);
 char* to_ascii(char *, int *);
 char* from_ascii(char *, int *);
-
+const char* get_errstr(void);
+void set_errstr(const char*);
 
 /*  from access.c */
 int read_access(const char*);
@@ -230,6 +279,7 @@ int add_to_options(char*);
 FILE* open_file(const char*);
 FILE* open_logfile(const char*);
 int read_config(const char*);
+int set_conffilename(const char*);
 float get_rate(const char*, const char*, unsigned long int,
 		const char*, unsigned long int, const char*);
 const char* get_sshpasswd(unsigned long int, const char*);
@@ -254,4 +304,4 @@ int gethostaddr(const char*, struct in_addr*);
 struct limitstruct;
 void free_limitst(struct limitstruct*);
 void free_loginst();
-
+void free_errstr();
